import logging

from cythonbuilder import appsettings
from cythonbuilder.logs import create_logger, LoggerSettings



# Logger
loggername = appsettings.package_name
loggerSettings = LoggerSettings(format=f'[%(name)s] %(message)s', level=logging.INFO)
loggerSettings.set_level_style(level='debug', color='green')
loggerSettings.set_level_style(level='info', color='white')
logger = create_logger(name=loggername, settings=loggerSettings)



import sys
import typing

import typer as typer

from cythonbuilder import cython_builder
from cythonbuilder import appsettings
from cythonbuilder import helpers
from cythonbuilder import set_logger_debug_mode
from cythonbuilder.definitions import DefaultArgs
from cythonbuilder.services import logger

_VERBOSE = '-v' in sys.argv or '--verbose' in sys.argv
app = typer.Typer(name="main_helpy", pretty_exceptions_enable=_VERBOSE)


@app.command(name="init", help=f"Initialize cythonbuilder", short_help="Initialize Cython Builder")
def init_cb():
    logger.info(msg=f"Initializing..")
    cython_builder.cy_init()
    logger.info(msg=f"Initialized")
    sys.exit(0)


@app.command(name="list", help="List all .pyx files", short_help="List all .pyx files")
def list_(
        target_filenames: typing.List[str] = typer.Option(None, "--files", help="Filter by these files"),
):
    # Lists all pyx files that can be built
    filtermsg = f" (filtered by {' '.join(target_filenames)}) " if (target_filenames) else ""
    logger.info(msg=f"Listing all pyx files{filtermsg}..")
    found_pyx_files: [str] = cython_builder.cy_list(target_files=target_filenames)
    if (len(found_pyx_files) > 0):
        found_files_string = "\n".join([f"\t - {fle}" for fle in found_pyx_files])
        logger.info(msg=f"Found files:\n{found_files_string}")
    else:
        logger.info(msg=f"No pyx files found")


@app.command(name="build", help="compile all .pyx files", short_help="Compile all .pyx files to C")
def build(
        target_filenames: typing.List[str] = typer.Option(None, "--files", help="Target .pyx file names"),
        include_numpy: bool = typer.Option(False, "--include-numpy", help="Include numpy if numpy is installed in your project"),
        dont_generate_annotations: bool = typer.Option(False, "--no-annotation", help="Skip generating annotations (html)"),
        dont_generate_pyi: bool = typer.Option(False, "--no-interface", help="Skip generating .pyi stub files"),
        keep_c_files: bool = typer.Option(False, "--no-cleanup", help="Skip removing generated C-files"),
        encoding: str = typer.Option('UTF-8', "--encoding", help="Encoding of your .pyx files"),
        ACCEPT: bool = DefaultArgs.accept
        # VERBOSE: bool = DefaultArgs.verbose, OVERWRITE: bool = DefaultArgs.overwrite, ACCEPT: bool = DefaultArgs.accept
):
    # Validate arguments
    numpy_is_installed = helpers.package_is_installed(package_import_name='numpy')
    if (include_numpy and not numpy_is_installed):
        if (input(f"[{appsettings.package_name}] You want to include numpy but your current project does not have numpy installed. Install numpy? (y/n)").lower() == 'y'):
            logger.warning(msg=f"Please install numpy with [pip install numpy]")
            sys.exit(0)
        else:
            include_numpy = False
    if (not include_numpy and numpy_is_installed):
        if (input(f"[{appsettings.package_name}] Your current project uses numpy. Do you want to include numpy in your Cython build? (y/n)").lower() == 'y'):
            include_numpy = True

    # Building
    logger.info(msg=f"Building Cython files..")
    try:
        # 1. Find pyx files
        found_pyx_files: [str] = cython_builder.cy_list(target_files=target_filenames)
        if (len(found_pyx_files) == 0):
            logger.info(msg="No .pyx files found")
            sys.exit(0)
        logger.debug(msg=f"Found {len(found_pyx_files)} to build")

        # 2. Confirm that we want to build
        if (not ACCEPT):
            __formatted_package_list = "\n".join(f"\t - {file}" for file in found_pyx_files)
            if (input(f"[{appsettings.package_name}] these {len(found_pyx_files)} pyx files?\n(y/n) \n {__formatted_package_list}").lower() != "y"):
                logger.info(msg="Exiting..", )
                sys.exit(0)
        logger.debug(msg=f"Building {len(found_pyx_files)} pyx files..")

        # 3. Build
        cython_builder.cy_build(
            target_files=found_pyx_files,
            create_annotations=not dont_generate_annotations,
            include_numpy=include_numpy,
        )
        logger.debug(msg=f"Built {len(found_pyx_files)} pyx files, cleaning up..")

        # 4. Cleanup after build
        cython_builder.cy_clean(target_files=found_pyx_files, keep_c_files=keep_c_files)
        logger.debug(msg=f"Cleanup complete")

        # 5.  Generate pyi files
        if (not dont_generate_pyi):
            logger.debug(msg=f"Generating interface files..")
            cython_builder.cy_interface(target_files=found_pyx_files, encoding=encoding)
            logger.debug(msg=f"Generated pyi files")

        logger.info(msg=f"Cython build success")
    except Exception as e:
        logger.error(msg=f"build error: {e}")
        sys.exit(1)

@app.command(name="clean", help="Clean your project; remove all .c ", short_help="Clean your project")
def cb_clean(
        target_filenames: typing.List[str] = typer.Option(None, "--files", help="Target .pyx file names"),
        keep_c_files: bool = typer.Option(False, "--no-cleanup", help="Skip removing generated C-files"),
        ACCEPT: bool = DefaultArgs.accept
        # VERBOSE: bool = DefaultArgs.verbose, OVERWRITE: bool = DefaultArgs.overwrite, ACCEPT: bool = DefaultArgs.accept
):
    # Clean
    logger.info(msg=f"Cleaning Cython files..")
    try:
        # 1. Find pyx files
        found_pyx_files: [str] = cython_builder.cy_list(target_files=target_filenames)
        if (len(found_pyx_files) == 0):
            logger.info(msg="No .pyx files found")
            sys.exit(0)
        logger.debug(msg=f"Found {len(found_pyx_files)} to clean")

        # 2. Confirm that we want to build
        if (not ACCEPT):
            __formatted_package_list = "\n".join(f"\t - {file}" for file in found_pyx_files)
            if (input(f"[{appsettings.package_name}] these {len(found_pyx_files)} pyx files?\n(y/n) \n {__formatted_package_list}").lower() != "y"):
                logger.info(msg="Exiting..", )
                sys.exit(0)
        logger.debug(msg=f"Cleaning {len(found_pyx_files)} pyx files..")

        # 3. Clean
        cython_builder.cy_clean(target_files=found_pyx_files, keep_c_files=keep_c_files)
        logger.debug(msg=f"Cleanup complete")
        logger.info(msg=f"Cython build success")
    except Exception as e:
        logger.error(msg=f"build error: {e}")
        sys.exit(1)

@app.command(name="interface", help="Create .pyi files for use in your Python project", short_help="generate interface files")
def cb_interface(
        target_filenames: typing.List[str] = typer.Option(None, "--files", help="Target .pyx file names"),
        ACCEPT: bool = DefaultArgs.accept
):
    logger.info(msg=f"Generating interface files..")
    try:
        # 1. Find pyx files
        found_pyx_files: [str] = cython_builder.cy_list(target_files=target_filenames)
        if (len(found_pyx_files) == 0):
            logger.info(msg="No .pyx files found")
            sys.exit(0)
        logger.debug(msg=f"Found {len(found_pyx_files)} to create pyi files for")

        # 2. Confirm that we want to build
        if (not ACCEPT):
            __formatted_package_list = "\n".join(f"\t - {file}" for file in found_pyx_files)
            if (input(f"[{appsettings.package_name}] these {len(found_pyx_files)} pyx files?\n(y/n) \n {__formatted_package_list}").lower() != "y"):
                logger.info(msg="Exiting..", )
                sys.exit(0)
        logger.debug(msg=f"Creating {len(found_pyx_files)} pyi files..")

        # 3. Generate pyi files
        logger.debug(msg=f"Generating pyi files..")
        cython_builder.cy_interface(target_files=found_pyx_files)
        logger.debug(msg=f"Generated pyi files")
        logger.info(msg=f"Generating interface files success")
    except Exception as e:
        logger.error(msg=f"build error: {e}")
        sys.exit(1)


def display_help():
    print(f"""{appsettings.package_name}
        Automatically builds and packages your Cython code 
        1. Initialize {appsettings.package_name} with `{appsettings.console_name} init` 
        3. Call `{appsettings.console_name} build` to build and package all .pyx files in your project 

        Commands:
        (call either command with -v to display more information)
        help        Show this screen
        init        Create 'ext' and 'ext/annotations' folders
        list        Lists all pyx files in your project that can be built
                    Pass a space-separated list of filenames (with or without .pyx) to filter
        build       Finds and builds all pyx files in your project (excluding venv directories) package cython files. Then cleans the project (see cybuilder clean)
                    Pass a space-separated list of filenames (with or without .pyx) to filter.
                    Options:
                      --no-interfaces   Prevents creating .pyi interface files (default False)
                      --include-numpy   Includes numpy (default False)
                      --no-annotation   Disables generating the annotations html (default True)
                      --no-cleanup      Prevents removal of intermediate C files generated by Cython (default True)
        clean       Moves all generated .html files to ext/annotations 
                    Pass a space-separated list of filenames (with or without .pyx) to filter.
                    Options:
                      --keep-c      Prevents removal of intermediate C files generated by Cython (default False)
        interface   Generates all pyi files that serve as an interface for IDE's
                    Pass a space-separated list of filenames (with or without .pyx) to filter.
    """)
    sys.exit(0)


def main():
    """ Parse command line """


    args = sys.argv[1:]

    # Preferences
    # FORCE = '-f' in args or '--force' in args
    VERBOSE = '-v' in args or '--verbose' in args
    ACCEPT = '-y' in args or '--yes' in args
    # OVERWRITE = '-o' in args or '--overwrite' in args
    for flag in ['-f', '--force', '-v', '--verbose', '-y', '--accept', '-o', '--overwrite']:
        args.remove(flag) if (flag in args) else None

    return
    # Logging level
    if (VERBOSE):
        set_logger_debug_mode(logger=logger)
        # from cythonbuilder.logs import set_format
        # set_format(logger=logger, format=f"[%(name)s] %(asctime)s %(module)-8s %(lineno)-3d  %(message)s")
        # logger.setLevel(logging.DEBUG)



    # Parse arguments - Must provide at least one argument
    if (len(args) == 0):
        display_help()
    cmd1 = helpers.CliTools.pop_arg_or_exit(arglist=args, errormessage=f"[{appsettings.package_name}] {appsettings.package_name} expects at least one argument. Check out [helpy help] for more information")
    logger.debug(msg=f"{cmd1=}")


    # HELPY functions
    if (cmd1 == 'ok'):
        pass
    # elif (cmd1 == 'init'):
    #     logger.info(msg=f"[{cmd1}] - Initializing..")
    #     cython_builder.cy_init()
    #     logger.info(msg=f"[{cmd1}] - Initialized")
    #     sys.exit(0)
    # elif (cmd1 == 'list'):
    #     # Lists all pyx files that can be built
    #     target_filenames = args if (len(args) > 0) else None
    #     filtermsg = f" (filtered by {' '.join(target_filenames)}) " if (target_filenames) else ""
    #     logger.info(msg=f"Listing all pyx files{filtermsg}..")
    #     found_pyx_files: [str] = cython_builder.cy_list(target_files=target_filenames)
    #     if (len(found_pyx_files) >0):
    #         found_files_string = "\n".join([f"\t - {fle}" for fle in found_pyx_files])
    #         logger.info(msg=f"Found files:\n{found_files_string}")
    #     else:
    #         logger.info(msg=f"No pyx files found")
    # elif (cmd1 == 'build'):
    #     # Get arguments
    #     include_numpy = '--include-numpy' in args
    #     dont_generate_annotations = '--no-annotations' in args
    #     dont_generate_pyi = '--no-interfaces' in args
    #     keep_c_files = '--no-cleanup' in args
    #     encoding = 'UTF-8'
    #     args = [a for a in args if (a not in ['--no-annotations', '--include-numpy', '--no-cleanup', '--no-interfaces'])]
    #     target_filenames = args if (len(args) > 0) else None
    #
    #     # Validate arguments
    #     numpy_is_installed = helpers.package_is_installed(package_import_name='numpy')
    #     if (include_numpy and not numpy_is_installed):
    #         if (input(f"[{appsettings.package_name} {cmd1}] You want to include numpy but your current project does not have numpy installed. Install numpy? (y/n)").lower() == 'y'):
    #             logger.warning(msg=f"Please install numpy with [pip install numpy]")
    #             sys.exit(0)
    #         else:
    #             include_numpy = False
    #     if (not include_numpy and numpy_is_installed):
    #         if (input(f"[{appsettings.package_name} {cmd1}] Your current project uses numpy. Do you want to include numpy in your Cython build? (y/n)").lower() == 'y'):
    #             include_numpy = True
    #
    #     logger.info(msg=f"[{cmd1}] - Building Cython files..")
    #     try:
    #         # 1. Find pyx files
    #         found_pyx_files: [str] = cython_builder.cy_list(target_files=target_filenames)
    #         if (len(found_pyx_files) == 0):
    #             logger.info(msg="No .pyx files found")
    #             sys.exit(0)
    #         logger.debug(msg=f"Found {len(found_pyx_files)} to build")
    #
    #         # 2. Confirm that we want to build
    #         if (not ACCEPT):
    #             __formatted_package_list = "\n".join(f"\t - {file}" for file in found_pyx_files)
    #             if (input(f"[{appsettings.package_name} {cmd1}] these {len(found_pyx_files)} pyx files?\n(y/n) \n {__formatted_package_list}").lower() != "y"):
    #                 logger.info(msg="Exiting..", )
    #                 sys.exit(0)
    #         logger.debug(msg=f"Building {len(found_pyx_files)} pyx files..")
    #
    #         # 3. Build
    #         cython_builder.cy_build(
    #             target_files=found_pyx_files,
    #             create_annotations=not dont_generate_annotations,
    #             include_numpy=include_numpy,
    #         )
    #         logger.debug(msg=f"Built {len(found_pyx_files)} pyx files, cleaning up..")
    #
    #         # 4. Cleanup after build
    #         cython_builder.cy_clean(target_files=found_pyx_files, keep_c_files=keep_c_files)
    #         logger.debug(msg=f"Cleanup complete")
    #
    #         # 5.  Generate pyi files
    #         if (not dont_generate_pyi):
    #             logger.debug(msg=f"Generating interface files..")
    #             cython_builder.cy_interface(target_files=found_pyx_files, encoding='zzz')
    #             logger.debug(msg=f"Generated pyi files")
    #
    #         logger.info(msg=f"[{cmd1}] - Cython build success")
    #     except Exception as e:
    #         logger.error(msg=f"[{cmd1}] build error: {e}")
    #         sys.exit(1)
    # elif (cmd1 == 'clean'):
    #     # Get arguments
    #     keep_c_files = '--keep-c' in args
    #     args = [a for a in args if (a not in ['--keep-c'])]
    #     target_filenames = args if (len(args) > 0) else None
    #
    #     # Clean
    #     logger.info(msg=f"[{cmd1}] - Cleaning Cython files..")
    #     try:
    #         # 1. Find pyx files
    #         found_pyx_files: [str] = cython_builder.cy_list(target_files=target_filenames)
    #         if (len(found_pyx_files) == 0):
    #             logger.info(msg="No .pyx files found")
    #             sys.exit(0)
    #         logger.debug(msg=f"Found {len(found_pyx_files)} to clean")
    #
    #         # 2. Confirm that we want to build
    #         if (not ACCEPT):
    #             __formatted_package_list = "\n".join(f"\t - {file}" for file in found_pyx_files)
    #             if (input(f"[{appsettings.package_name} {cmd1}] these {len(found_pyx_files)} pyx files?\n(y/n) \n {__formatted_package_list}").lower() != "y"):
    #                 logger.info(msg="Exiting..", )
    #                 sys.exit(0)
    #         logger.debug(msg=f"Cleaning {len(found_pyx_files)} pyx files..")
    #
    #         # 3. Clean
    #         cython_builder.cy_clean(target_files=found_pyx_files, keep_c_files=keep_c_files)
    #         logger.debug(msg=f"Cleanup complete")
    #         logger.info(msg=f"[{cmd1}] - Cython build success")
    #     except Exception as e:
    #         logger.error(msg=f"[{cmd1}] build error: {e}")
    #         sys.exit(1)
    elif (cmd1 in ['interface', 'pyi']):

        target_filenames = args if (len(args) > 0) else None

        logger.info(msg=f"[{cmd1}] - Generating interface files..")
        try:
            # 1. Find pyx files
            found_pyx_files: [str] = cython_builder.cy_list(target_files=target_filenames)
            if (len(found_pyx_files) == 0):
                logger.info(msg="No .pyx files found")
                sys.exit(0)
            logger.debug(msg=f"Found {len(found_pyx_files)} to create pyi files for")

            # 2. Confirm that we want to build
            if (not ACCEPT):
                __formatted_package_list = "\n".join(f"\t - {file}" for file in found_pyx_files)
                if (input(f"[{appsettings.package_name} {cmd1}] these {len(found_pyx_files)} pyx files?\n(y/n) \n {__formatted_package_list}").lower() != "y"):
                    logger.info(msg="Exiting..", )
                    sys.exit(0)
            logger.debug(msg=f"Creating {len(found_pyx_files)} pyi files..")

            # 3. Generate pyi files
            logger.debug(msg=f"Generating pyi files..")
            cython_builder.cy_interface(target_files=found_pyx_files)
            logger.debug(msg=f"Generated pyi files")
            logger.info(msg=f"[{cmd1}] - Generating interface files success")
        except Exception as e:
            logger.error(msg=f"[{cmd1}] build error: {e}")
            sys.exit(1)

    elif (cmd1 == 'help'):
        display_help()
    else:
        logger.warning(msg=f"unknown command: '{cmd1}'")
        display_help()




import os
import sys
from dataclasses import dataclass

import typer

from cythonbuilder import appsettings

ROOT_DIR = os.path.realpath(os.path.join(os.path.dirname(__file__), '..'))
PACKAGE_ROOT = sys.modules[appsettings.package_name].__path__[0]

@dataclass
class DefaultArgs:
    verbose:bool =   typer.Option(False, "--verbose", '-v', help="Higher verbosity of commands")
    force:bool =     typer.Option(False, "--force", '-f', help="Force this operation")#, prompt=f"{helpy_mascotte} Are you sure you want to delete the user?", )
    overwrite:bool = typer.Option(False, "--overwrite", '-o', help="Overwrite existing")
    accept:bool =    typer.Option(False, "--accept", '-y', help="Yes to all prompts")



# import code from your package here to make them available outside the module like:
from cythonbuilder.cython_builder import cy_init, cy_clean, cy_build, cy_list, cy_interface
from cythonbuilder.logs import LoggerSettings, create_logger, set_logger_debug_mode

__version__ = "0.1.14"
__last_build_datetime__ ="2023-03-25 13:05:17"
__last_publish_version__ ="0.1.13"
__last_publish_datetime__ ="2022-06-01 12:08:36"




import io
import re
from dataclasses import dataclass
from typing import TextIO

from cythonbuilder.services import logger

class LineConverter:

    in_class_body:bool=False
    in_func_body:bool=False

    def __init__(self, line:str, file_spaces_for_one_tab:int=4):
        self.pyx_line = line
        self.file_spaces_for_one_tab = file_spaces_for_one_tab

    # Indentation
    @property
    def indent_spacecount(self) -> int:
        """ How many spaces is the provided line indented? """
        spaces: int = 0
        if (len(self.pyx_line) > 0):
            if (self.pyx_line[0] == ' '):
                for letter_idx in range(len(self.pyx_line)):
                    if (self.pyx_line[letter_idx + 1] != ' '):
                        spaces = letter_idx + 1
                        break
        return spaces
    @property
    def indent_tabs(self) -> int:
        """ How many spaces is the provided line indented? """
        spaces = get_line_indentation_spacecount(line=self.pyx_line)
        return int(spaces / self.file_spaces_for_one_tab)

    # Definitions
    @property
    def is_import(self) -> bool:
        """ """
        line = self.pyx_line.strip()
        word_array = line.split(" ")
        first_word_is_import = word_array[0] in ['import', 'cimport']
        line_contains_from_import = word_array[0] == 'from' and 'import' in word_array
        return any([first_word_is_import, line_contains_from_import])
    @property
    def is_class_def(self) -> bool:
        """ """
        return self.pyx_line[-1] == ":" and "class " in self.pyx_line
    @property
    def is_func_def(self) -> bool:
        """ Is C or PY function """
        line = self.pyx_line.strip(' ')
        word_array = line.split()
        endswith_colon = line[-1] == ':'
        contains_def = len(set(word_array).intersection({'cdef', 'def', 'cpdef'})) > 0
        is_no_class = not self.is_class_def
        return all([endswith_colon, contains_def, is_no_class])
    @property
    def is_c_func_def(self) -> bool:
        """ Is function but also a C function """
        is_c_function = False
        if (self.is_func_def):
            line_parts_set = set(self.pyx_line.split(" "))
            cfunction_defs = {'cpdef', 'cdef'}
            if (len(line_parts_set.intersection(cfunction_defs)) > 0):
                is_c_function = True
        return is_c_function
    @property
    def is_property(self) -> bool:
        line = self.pyx_line.strip()
        is_property = line[0] == '@'
        is_not_cython_property = '@cython.' not in line
        return all([is_property, is_not_cython_property])
    @property
    def is_define_line(self) -> bool:
        """ This line is not an import, function, property or class def"""
        return any([self.is_import, self.is_class_def, self.is_func_def, self.is_property])

    # translate
    @property
    def py_line(self) -> str:

        # Strip and remove double spaces
        py_line = self.pyx_line.strip()
        py_line = " ".join(py_line.split("  "))
        if (self.is_import):
            py_line = py_line.replace("cimport", "import")
        if (self.is_class_def):
            py_line = " ".join([w for w in py_line.split(" ") if (w not in ['cdef', 'cpdef'])])
        if (self.is_func_def):

            # replace definition
            py_line = py_line.replace('cdef', 'def')
            py_line = py_line.replace('cpdef', 'def')

            # Handle return function types
            py_return_type = None
            if ('->' in py_line):
                py_return_type = py_line.split('->')[1].strip().replace(":", "")
                py_line = f"{py_line.split('->')[0].strip()}:"

            array_between_brackets = re.findall('\(.*?\)', py_line)
            logger.error(py_line)
            for brackets in array_between_brackets:
                # Replace C return type to -> returntype
                func_part_one_old = py_line.replace(f"{brackets}:", "")
                func_sig_parts = func_part_one_old.split(" ")

                if (len(func_sig_parts) == 3):
                    py_return_type = self.convert_type_cy_to_py(cy_type=func_sig_parts[1])
                func_part_one_new = " ".join([func_sig_parts[0], func_sig_parts[-1]])
                py_line = py_line.replace(func_part_one_old, func_part_one_new)
                for argument in brackets.split(","):

                    # Strip away parentheses and spaces
                    argument = argument.strip("() ")


                    # Rework arguments of c-function to py-style
                    if (self.is_c_func_def):
                        # C types are like (int age)
                        arg_name_py:str = None
                        arg_type_py:str = None
                        arg_default_value_py:str = None
                        # take out arg name and type if they are so specified
                        if (' ' in argument):
                            arg_name_py = argument.split(' ')[1]
                            arg_type_py = self.convert_type_cy_to_py(argument.split(' ')[0])
                        else:
                            arg_name_py = argument

                        # Get default value
                        if ('=' in argument):
                            arg_default_value_py = argument.split("=")[1].strip()

                        # make sure default value is not part of name
                        if (arg_name_py != None and "=" in arg_name_py):
                            arg_name_py = arg_name_py.split("=")[0]
                        arg_type_py = f":{arg_type_py}" if (arg_type_py != None) else ''
                        arg_default_value_py = f"={arg_default_value_py}" if (arg_default_value_py != None) else ''
                        newArg = f"{arg_name_py}{arg_type_py}{arg_default_value_py}"
                        py_line = py_line.replace(argument, newArg)
            # Add python return type
            if (py_return_type != None):
                py_line = py_line.strip(":") + f' -> {py_return_type}:'

            # Add ... to function content
            py_line = py_line + f"\n{self.file_spaces_for_one_tab * (self.indent_tabs + 1) * ' '}..."
        # return indentation * spaces_for_one_tab * ' ' + line

        # add original indentation back
        py_line = f"{self.file_spaces_for_one_tab * self.indent_tabs * ' '}" + py_line

        return py_line


    def convert_type_cy_to_py(self, cy_type:str):
        """ """
        # todo continue https://stackoverflow.com/questions/55451545/what-are-all-the-types-available-in-cython
        cy_type = cy_type.lower()

        if (cy_type in ['bint', 'bool']):
            return 'bool'
        elif (cy_type in ['char', 'short', 'int', 'long', 'long long']):
            return 'int'
        elif (cy_type in ['float', 'double', 'long double']):
            return 'float'
        elif (cy_type in ['float complex', 'double complex', 'complex']):
            return 'complex'
        elif (cy_type in ['char*', 'std::string', 'str']):
            return 'str'
        elif (cy_type in ['void']):
            return 'None'
        else:
            # non-built-in type like np.ndarray
            return cy_type




def pyx_to_pyi(open_pyx:TextIO) -> [str]:
    """ Reads all content from a pyx file, convers and writes a pyi """

    # 1. Read pyx file
    pyx_lines: [str] = open_pyx.readlines()

    # 2. Strip away any content after a #
    pyx_lines = list(map(lambda l: l.split("#")[0].rstrip() if "#" in l else l, pyx_lines))

    # 3. Skip empty lines
    pyx_lines = [l.strip("\n") for l in pyx_lines]
    pyx_lines = [l for l in pyx_lines if (len(l) > 0)]

    # 4. Determine file indentation
    all_indentations = [get_line_indentation_spacecount(line=l) for l in pyx_lines]
    all_indentations = [i for i in all_indentations if (i > 0)]
    spaces_for_one_tab = min(all_indentations)
    # Extra check: test wheter all are divisible by the space_for_one_tab
    for indent in all_indentations:
        if (indent % spaces_for_one_tab != 0):
            raise ValueError(f"Found invalid indentation: {indent} not divisible by {spaces_for_one_tab}")


    # 5. Convert lines to py
    py_lines:[str] = []
    prev_line:LineConverter = None
    for line_idx in range(len(pyx_lines)):
        pyxline = pyx_lines[line_idx]

        ld = LineConverter(line=pyxline, file_spaces_for_one_tab=spaces_for_one_tab)

        if (prev_line != None):
            # Classes
            if (prev_line.is_class_def):
                ld.in_class_body = True
            elif (prev_line.in_class_body and not ld.is_class_def):
                ld.in_class_body = True

            # Functions
            if (prev_line.is_func_def):
                ld.in_func_body = True
            elif (prev_line.in_func_body and (not ld.is_class_def and not ld.is_func_def)):
                ld.in_func_body = True

            # if current line defines anything
            if (ld.is_define_line):
                ld.in_func_body = False
                ld.in_class_body = False

            # kep define lines an any lines that are in a class_body
            if (not ld.in_class_body and not ld.is_define_line):
                # logger.error(msg=f"{prev_line.indent_tabs},{ld.indent_tabs} - {l_is_class}{l_is_func}-{'C' if ld.in_class_body else 'X'}{'F' if ld.in_func_body else 'X'}, {ld.is_property} {pyxline}")
                continue
            # else:
            # logger.debug(msg=f"{prev_line.indent_tabs},{ld.indent_tabs} - {l_is_class}{l_is_func}-{'C' if ld.in_class_body else 'X'}{'F' if ld.in_func_body else 'X'}, {ld.is_property} {pyxline}")

            # logger.warning(msg=f"{ld.pyx_line}")
            # logger.debug(msg=f"{ld.py_line}")


        prev_line = ld
        py_lines.append(ld.py_line)

    # 6. return
    return [f"{line}\n" for line in py_lines]
    # with open(target_pyi_path, 'w') as pyi_out:
    #     pyi_out.writelines([f"{line}\n" for line in py_lines])





def get_line_indentation_spacecount(line: str) -> int:
    """ How many spaces is the provided line indented? """
    spaces: int = 0
    if (len(line) > 0):
        if (line[0] == ' '):
            for letter_idx in range(len(line)):
                if (line[letter_idx + 1] != ' '):
                    spaces = letter_idx + 1
                    break
    return spaces
def get_line_indentation(line: str, file_spaces_for_one_tab:int) -> int:
    """ How many spaces is the provided line indented? """
    spaces = get_line_indentation_spacecount(line=line)
    return int(spaces / file_spaces_for_one_tab)